@@ -1,4 +1,4 @@
-__all__ = ['Nlp', 'MultistartNlp']
+__all__ = ["Nlp", "MultistartNlp"]
 
-from csnlp.nlp import Nlp
 from csnlp.multistart_nlp import MultistartNlp
+from csnlp.nlp import Nlp
