@@ -1,18 +1,27 @@
-from functools import lru_cache
-from typing import Any, Dict, Iterable, List, Literal, Optional, Tuple, Union
+from contextlib import contextmanager
+from functools import lru_cache, partial
+from typing import (
+    Any,
+    Dict,
+    Generic,
+    Iterable,
+    List,
+    Literal,
+    Optional,
+    Tuple,
+    TypeVar,
+    Union,
+)
 
+import casadi as cs
 import numpy as np
 
-from csnlp.nlp import (
-    DMStruct,
-    Nlp,
-    Solution,
-    cache_clearer,
-    cs,
-    dict2struct,
-    partial,
-    subsevalf,
-)
+from csnlp.nlp import Nlp
+from csnlp.solutions import Solution, subsevalf
+from csnlp.util.data import DMStruct, dict2struct
+from csnlp.util.funcs import invalidate_cache
+
+T = TypeVar("T", cs.SX, cs.MX)
 
 
 def _n(sym_name: str, scenario: int) -> str:
@@ -25,12 +34,14 @@ def _get_value(x, sol: Solution, old, new, eval: bool = True):
     return sol._get_value(cs.substitute(x, old, new), eval=eval)
 
 
-class MultistartNlp(Nlp):
+class MultistartNlp(Nlp[T], Generic[T]):
     """A class to easily model and solve an NLP from multiple starting
     initial guesses in parallel. This is especially useful in case of strong
     nonlinearities, where the solver's initial conditions play a great role in
     the optimality of the solution (rather, its sub-optimality)."""
 
+    is_multi: bool = True
+
     def __init__(
         self,
         *args,
@@ -61,7 +72,17 @@ class MultistartNlp(Nlp):
         # overridden to create multiples of these in the hidden nlp
         super().__init__(*args, **kwargs)
         self._multi_nlp = Nlp(*args, **kwargs)  # actual nlp
-        self._vars_per_start: Dict[int, Dict[str, Union[cs.SX, cs.MX]]] = {}
+        self._vars_per_start: Dict[int, Dict[str, T]] = {}
+
+    @contextmanager
+    def fullstate(self) -> None:
+        with super().fullstate(), self._multi_nlp.fullstate():
+            yield
+
+    @contextmanager
+    def pickleable(self) -> None:
+        with super().pickleable(), self._multi_nlp.pickleable():
+            yield
 
     @property
     def starts(self) -> int:
@@ -75,66 +96,62 @@ class MultistartNlp(Nlp):
         vars: bool = False,
         pars: bool = False,
         dual: bool = False,
-    ) -> Dict[str, Union[cs.SX, cs.MX]]:
+    ) -> Dict[str, T]:
         """Internal utility to retrieve the symbols of the i-th scenario."""
-        S: Dict[str, Union[cs.SX, cs.MX]] = {}
+        S: Dict[str, T] = {}
         if vars:
             S.update(
                 self._vars
                 if i is None
-                else {k: self._multi_nlp._vars[_n(k, i)] for k in self._vars}
+                else {k: self._multi_nlp.unwrapped._vars[_n(k, i)] for k in self._vars}
             )
         if pars:
             S.update(
                 self._pars
                 if i is None
-                else {k: self._multi_nlp._pars[_n(k, i)] for k in self._pars}
+                else {k: self._multi_nlp.unwrapped._pars[_n(k, i)] for k in self._pars}
             )
         if dual:
             S.update(
                 self._dual_vars
                 if i is None
-                else {k: self._multi_nlp._dual_vars[_n(k, i)] for k in self._dual_vars}
+                else {
+                    k: self._multi_nlp.unwrapped._dual_vars[_n(k, i)]
+                    for k in self._dual_vars
+                }
             )
         return S
 
-    @cache_clearer(_symbols)
-    def parameter(
-        self, name: str, shape: Tuple[int, int] = (1, 1)
-    ) -> Union[cs.SX, cs.MX]:
+    @invalidate_cache(_symbols)
+    def parameter(self, name: str, shape: Tuple[int, int] = (1, 1)) -> T:
         out = super().parameter(name, shape)
         for i in range(self._starts):
             self._multi_nlp.parameter(_n(name, i), shape)
         return out
 
-    @cache_clearer(_symbols)
+    @invalidate_cache(_symbols)
     def variable(
         self,
         name: str,
         shape: Tuple[int, int] = (1, 1),
         lb: Union[np.ndarray, cs.DM] = -np.inf,
         ub: Union[np.ndarray, cs.DM] = +np.inf,
-    ) -> Union[Tuple[cs.SX, cs.SX, cs.SX], Tuple[cs.MX, cs.MX, cs.MX]]:
+    ) -> Tuple[T, T, T]:
         out = super().variable(name, shape, lb, ub)
         for i in range(self._starts):
             self._multi_nlp.variable(_n(name, i), shape, lb, ub)
         return out
 
-    @cache_clearer(_symbols)
+    @invalidate_cache(_symbols)
     def constraint(
         self,
         name: str,
-        lhs: Union[np.ndarray, cs.DM, cs.SX, cs.MX],
+        lhs: Union[T, np.ndarray, cs.DM],
         op: Literal["==", ">=", "<="],
-        rhs: Union[np.ndarray, cs.DM, cs.SX, cs.MX],
+        rhs: Union[T, np.ndarray, cs.DM],
         soft: bool = False,
         simplify: bool = True,
-    ) -> Union[
-        Tuple[cs.SX, cs.SX],
-        Tuple[cs.MX, cs.MX],
-        Tuple[cs.SX, cs.SX, cs.SX],
-        Tuple[cs.MX, cs.MX, cs.MX],
-    ]:
+    ) -> Union[Tuple[T, T], Tuple[T, T, T]]:
         expr = lhs - rhs
         if simplify:
             expr = cs.simplify(expr)
@@ -143,14 +160,14 @@ class MultistartNlp(Nlp):
         symbols = self._symbols(vars=True, pars=True)
         for i in range(self._starts):
             symbols_i = self._symbols(i, vars=True, pars=True)
-            expr = subsevalf(expr, symbols, symbols_i, eval=False)
-            self._multi_nlp.constraint(_n(name, i), expr, op, 0, soft, simplify=False)
+            expr_i = subsevalf(expr, symbols, symbols_i, eval=False)
+            self._multi_nlp.constraint(_n(name, i), expr_i, op, 0, soft, simplify=False)
         return out
 
-    def minimize(self, objective: Union[cs.SX, cs.MX]) -> None:
+    def minimize(self, objective: T) -> None:
         out = super().minimize(objective)
         symbols = self._symbols(vars=True, pars=True)
-        self._fs: Union[List[cs.SX], List[cs.MX]] = [
+        self._fs: List[T] = [
             subsevalf(
                 objective, symbols, self._symbols(i, vars=True, pars=True), eval=False
             )
@@ -170,7 +187,7 @@ class MultistartNlp(Nlp):
         vals0: Union[None, Iterable[DMStruct], Iterable[Dict[str, np.ndarray]]] = None,
         return_all_sols: bool = False,
         return_multi_sol: bool = False,
-    ) -> Union[Solution, List[Solution]]:
+    ) -> Union[Solution[T], List[Solution[T]]]:
         """Solves the NLP with multiple initial conditions.
 
         Parameters
@@ -234,7 +251,7 @@ class MultistartNlp(Nlp):
             self._p, self._x, self._lam_g, self._lam_h, self._lam_lbx, self._lam_ubx
         )
 
-        sols: List[Solution] = []
+        sols: List[Solution[T]] = []
         fs = [float(multi_sol.value(f)) for f in self._fs]
         idx = range(self._starts) if return_all_sols else (np.argmin(fs),)
         for i in idx:
@@ -245,7 +262,7 @@ class MultistartNlp(Nlp):
             get_value = partial(_get_value, sol=multi_sol, old=old, new=new)
 
             sols.append(
-                Solution(
+                Solution[T](
                     f=fs[i],
                     vars=vars_,
                     vals=dict2struct(vals),
@@ -254,3 +271,6 @@ class MultistartNlp(Nlp):
                 )
             )
         return sols if return_all_sols else sols[0]
+
+    def __call__(self, *args, **kwargs):
+        return self.solve_multi(*args, **kwargs)
