@@ -1,30 +1,31 @@
-from typing import Dict, Optional, Tuple, Union
+from typing import Dict, Optional, Tuple, TypeVar, Union
 
 import casadi as cs
 import numpy as np
 
-from csnlp.nlp import _DUAL_VARIABLES_ORDER
 from csnlp.solutions import Solution
 from csnlp.util.array import hohessian, hojacobian
 from csnlp.util.data import array2cs, cs2array
-from csnlp.util.funcs import cache_clearer, cached_property
+from csnlp.util.funcs import cached_property, invalidate_cache
 from csnlp.wrappers.wrapper import Nlp, Wrapper
 
+T = TypeVar("T", cs.SX, cs.MX)
 
-class NlpSensitivity(Wrapper):
+
+class NlpSensitivity(Wrapper[T]):
     """
-    Wraps an NLP to allow to perform numerical sensitivity analysis and compute
-    its derivates. See [1] for nonlinear programming sensitivity analysis.
+    Wraps an NLP to allow to perform numerical sensitivity analysis and compute its
+    derivates. See [1] for nonlinear programming sensitivity analysis.
 
     References
     ----------
     [1] Buskens, C. and Maurer, H. (2001). Sensitivity analysis and real-time
-        optimization of parametric nonlinear programming problems. In M.
-        Grotschel, S.O. Krumke, and J. Rambau (eds.), Online Optimization of
-        Large Scale Systems, 3–16. Springer, Berlin, Heidelberg.
+        optimization of parametric nonlinear programming problems. In M. Grotschel, S.O.
+        Krumke, and J. Rambau (eds.), Online Optimization of Large Scale Systems, 3–16.
+        Springer, Berlin, Heidelberg.
     """
 
-    def __init__(self, nlp: Nlp, include_barrier_term: bool = True) -> None:
+    def __init__(self, nlp: Nlp[T], include_barrier_term: bool = True) -> None:
         """Instantiates the wrapper for performing NLP sensitivities.
 
         Parameters
@@ -33,16 +34,16 @@ class NlpSensitivity(Wrapper):
             The NLP problem to be wrapped.
         include_barrier_term : bool, optional
             If `True`, includes in the KKT matrix a new symbolic variable that
-            represents the barrier function of the interior-point solver.
-            Otherwise, no additional variable is added. See property `kkt` for
-            more details. By default `True`.
+            represents the barrier function of the interior-point solver. Otherwise, no
+            additional variable is added. See property `kkt` for more details. By
+            default `True`.
         """
         super().__init__(nlp)
         self.include_barrier_term = include_barrier_term
-        self._tau = self.nlp._csXX.sym("tau") if include_barrier_term else 0
+        self._tau = nlp.sym_type.sym("tau") if include_barrier_term else 0
 
     @cached_property
-    def lagrangian(self) -> Union[cs.SX, cs.MX]:
+    def lagrangian(self) -> T:
         """Gets the Lagrangian of the NLP problem (usually, `L`)."""
         h_lbx, lam_h_lbx = self.nlp.h_lbx
         h_ubx, lam_h_ubx = self.nlp.h_ubx
@@ -55,30 +56,27 @@ class NlpSensitivity(Wrapper):
         )
 
     @cached_property
-    def kkt(
-        self,
-    ) -> Union[Tuple[cs.SX, Optional[cs.SX]], Tuple[cs.MX, Optional[cs.MX]]]:
+    def kkt(self) -> Tuple[T, Optional[T]]:
         """Gets the KKT conditions of the NLP problem in vector form, i.e.,
         ```
                             |      dLdx     |
                         K = |       G       |
                             | diag(lam_h)*H |
         ```
-        where `dLdx` is the gradient of the lagrangian w.r.t. the primal
-        variables `x`, `G` collects the equality constraints, `H` collects
-        the inequality constraints and `lam_h` its corresponding dual
-        variables.
+        where `dLdx` is the gradient of the lagrangian w.r.t. the primal variables `x`,
+        `G` collects the equality constraints, `H` collects the inequality constraints
+        and `lam_h` its corresponding dual variables.
 
-        If `include_barrier_term=True`, the inequalities include an additional
-        barrier term `tau`, so that
+        If `include_barrier_term=True`, the inequalities include an additional barrier
+        term `tau`, so that
         ```
                             diag(lam_h)*H + tau
         ```
-        which is also returned as the second element of the tuple. Otherwise,
-        `tau` is `None`.
+        which is also returned as the second element of the tuple. Otherwise, `tau` is
+        `None`.
 
-        Note: The order of the KKT conditions can be adjusted via the
-        `_PRIMAL_DUAL_ORDER` dict.
+        Note: The order of the KKT conditions can be adjusted via the class attribute
+        `_PRIMAL_DUAL_ORDER`.
         """
         items = {
             "g": self.nlp._g,
@@ -88,15 +86,15 @@ class NlpSensitivity(Wrapper):
         }
         kkt = cs.vertcat(
             cs.jacobian(self.lagrangian, self.nlp._x).T,
-            *(items.pop(v) for v in _DUAL_VARIABLES_ORDER)
+            *(items.pop(v) for v in self.nlp._DUAL_VARIABLES_ORDER)
         )
         assert not items, "Internal error. _DUAL_VARIABLES_ORDER modified."
         return kkt, (self._tau if self.include_barrier_term else None)
 
     @cached_property
-    def jacobians(self) -> Dict[str, Union[cs.SX, cs.MX]]:
-        """Computes various partial derivatives, which are then grouped in a
-        dict with the following entries
+    def jacobians(self) -> Dict[str, T]:
+        """Computes various partial derivatives, which are then grouped in a dict with
+        the following entries
             - `L-p`: lagrangian w.r.t. parameters
             - `g-x`: equality constraints w.r.t. primal variables
             - `h-x`: inequality constraints w.r.t. primal variables
@@ -114,9 +112,9 @@ class NlpSensitivity(Wrapper):
         }
 
     @cached_property
-    def hessians(self) -> Dict[str, Union[cs.SX, cs.MX]]:
-        """Computes various partial hessians, which are then grouped in a
-        dict with the following entries
+    def hessians(self) -> Dict[str, T]:
+        """Computes various partial hessians, which are then grouped in a dict with the
+        following entries
             - `L-pp`: lagrangian w.r.t. parameters (twice)
             - `L-xx`: lagrangian w.r.t. primal variables (twice)
             - `L-px`: lagrangian w.r.t. parameters and then primal variables
@@ -132,8 +130,8 @@ class NlpSensitivity(Wrapper):
 
     @cached_property
     def hojacobians(self) -> Dict[str, np.ndarray]:
-        """Computes various 3D jacobians, which are then grouped in a dict with
-        the following entries
+        """Computes various 3D jacobians, which are then grouped in a dict with the
+        following entries
             - `K-pp`: kkt conditions w.r.t. parameters (twice)
             - `K-yp`: kkt conditions w.r.t. parameters and primal variables
             - `K-yy`: kkt conditions w.r.t. primal variables (twice)
@@ -151,7 +149,7 @@ class NlpSensitivity(Wrapper):
         }
 
     @property
-    def licq(self) -> Union[np.ndarray, cs.SX, cs.MX]:
+    def licq(self) -> Union[np.ndarray, T]:
         """Gets the symbolic matrix for LICQ, defined as
         ```
                     LICQ = [ dgdx^T, dhdx^T ]^T
@@ -160,42 +158,40 @@ class NlpSensitivity(Wrapper):
         Independence Constraint Qualification.
 
         Note:
-            1) the LICQ are computed for only the active inenquality
-               constraints. Since this is a symbolic representation, all are
-               included, and it's up to the user to eliminate
-               the inactive.
+            1) the LICQ are computed for only the active inenquality constraints. Since
+               this is a symbolic representation, all are included, and it's up to the
+               user to eliminate the inactive.
 
-            2) lower and upper bound inequality constraints are not included in
-               `h` since they are by nature linear independent.
+            2) lower and upper bound inequality constraints are not included in `h`
+               since they are by nature linear independent.
         """
         return cs.vertcat(self.jacobians["g-x"], self.jacobians["h-x"])
 
     def parametric_sensitivity(
         self,
-        expr: Union[cs.SX, cs.MX] = None,
+        expr: T = None,
         solution: Optional[Solution] = None,
-    ) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[cs.SX, cs.SX], Tuple[cs.MX, cs.MX]]:
-        """Performs the (symbolic or numerical) sensitivity of the NLP w.r.t.
-        its parametrization, according to [1].
+    ) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[T, T]]:
+        """Performs the (symbolic or numerical) sensitivity of the NLP w.r.t. its
+        parametrization, according to [1].
 
         Parameters
         ----------
-        expression : cs.SX or MX, optional
-            If provided, computes the sensitivity of this expression (which
-            must be dependent on the primal-dual variables and/or parameters of
-            the NLP) w.r.t. the NLP parameters. If `None`, then the sensitivity
-            of the primal-dual variables is returned.
+        expression : casadi.SX or MX, optional
+            If provided, computes the sensitivity of this expression (which must be
+            dependent on the primal-dual variables and/or parameters of the NLP) w.r.t.
+            the NLP parameters. If `None`, then the sensitivity of the primal-dual
+            variables is returned.
         solution : Solution, optional
-            If a solution is passed, then the sensitivity is numerically
-            computed for that solution; otherwise, the sensitivity is carried
-            out symbolically (however, this is much more computationally
-            intensive).
+            If a solution is passed, then the sensitivity is numerically computed for
+            that solution; otherwise, the sensitivity is carried out symbolically
+            (however, this is much more computationally intensive).
 
         Returns
         -------
-        Union[np.ndarray, cs.SX, cs.MX]
-            The NLP parametric sensitivity in the form of an numerical array,
-            if a solution is passed, or a symbolic vector.
+        Union[np.ndarray, casadi.SX, casadi.MX]
+            The NLP parametric sensitivity in the form of an numerical array, if a
+            solution is passed, or a symbolic vector.
 
         Raises
         ------
@@ -204,11 +200,10 @@ class NlpSensitivity(Wrapper):
 
         References
         ----------
-        [1] Buskens, C. and Maurer, H. (2001). Sensitivity analysis and
-            real-time optimization of parametric nonlinear programming
-            problems. In M. Grotschel, S.O. Krumke, and J. Rambau (eds.),
-            Online Optimization of Large Scale Systems, 3–16. Springer, Berlin,
-            Heidelberg.
+        [1] Buskens, C. and Maurer, H. (2001). Sensitivity analysis and real-time
+            optimization of parametric nonlinear programming problems. In M. Grotschel,
+            S.O. Krumke, and J. Rambau (eds.), Online Optimization of Large Scale
+            Systems, 3–16. Springer, Berlin, Heidelberg.
         """
         # first order sensitivity, a.k.a., dydp
         Ky = self.jacobians["K-y"]
@@ -276,26 +271,26 @@ class NlpSensitivity(Wrapper):
             return solution.value(dzdp), solution.value(d2zdp2)
         return dzdp, d2zdp2
 
-    @cache_clearer(jacobians, hessians, hojacobians)
+    @invalidate_cache(jacobians, hessians, hojacobians)
     def parameter(self, *args, **kwargs):
         return self.nlp.parameter(*args, **kwargs)
 
-    @cache_clearer(lagrangian, kkt, jacobians, hessians, hojacobians)
+    @invalidate_cache(lagrangian, kkt, jacobians, hessians, hojacobians)
     def variable(self, *args, **kwargs):
         return self.nlp.variable(*args, **kwargs)
 
-    @cache_clearer(lagrangian, kkt, jacobians, hessians, hojacobians)
+    @invalidate_cache(lagrangian, kkt, jacobians, hessians, hojacobians)
     def constraint(self, *args, **kwargs):
         return self.nlp.constraint(*args, **kwargs)
 
-    @cache_clearer(lagrangian, kkt, jacobians, hessians, hojacobians)
+    @invalidate_cache(lagrangian, kkt, jacobians, hessians, hojacobians)
     def minimize(self, *args, **kwargs):
         return self.nlp.minimize(*args, **kwargs)
 
     @property
-    def _y_idx(self) -> Tuple[Union[cs.SX, cs.MX], Union[slice, np.ndarray]]:
-        """Internal utility to return all the primal-dual variables and indices
-        that are associated to non-redundant entries in the kkt conditions."""
+    def _y_idx(self) -> Tuple[T, Union[slice, np.ndarray]]:
+        """Internal utility to return all the primal-dual variables and indices that are
+        associated to non-redundant entries in the kkt conditions."""
         if self.nlp._csXX is cs.SX:
             y = self.nlp.primal_dual_vars()
             idx = slice(None)
